#include "axis.h"
#include <cmath>	// in w2k3ddk/inc/crt/, unbedingt vor plotxy.h
#include "../main.h"
#include "cplotxy.h"
#include <tchar.h>
#include <stdio.h>	// sntprintf() bei ANSI
#include <shlwapi.h>

/****************
 * DisplayRange *
 ****************/

int Axis::genLinTicks(DWORD flags, int intervals) {
 ticks.clear();	// Ruft Destruktoren auf
 ensureMin(intervals,1);
 Float dx=(dispRange.e-dispRange.a)/intervals;
 ensureMin(dx,dispRange.aufloes);
 dx=roundWithMethod(dx,flags);
 if (flags&1<<bIntegralB)
	dispRange.a=roundToMultiple(dispRange.a,-dx);	// glatt abrunden
 if (flags&1<<bIntegralE)
	dispRange.e=roundToMultiple(dispRange.e,dx);	// glatt aufrunden
 Float p = dispRange.a;
 if (flags&(1<<bTime|1<<bShowZ|1<<bCenterZ)) p=roundToMultiple(p,dx);
	// Im Fall der Zeitachse (fr Historie) Grenzen nicht erweitern
	// dadurch rollt die Skale bei voller Historie gleichmig
 ticks.reserve(intervals+1);
 int i=0;
 for (; i<=intervals; p+=dx) {
  if (!(flags&1<<bIntegralE) && p>dispRange.e) {
   if (flags&1<<bShowE) new(&ticks[i++]) Tick(*this,dispRange.e);
   break;
  }
  new(&ticks[i++]) Tick(*this,p);	// no string yet
  if (p>=dispRange.e) break;
 }
 if (flags&1<<bIntegralE) dispRange.e=p;
 ticks.commit_undefined(i);		// constructors already called, change _end internally
 return i;
}

// TODO: Was fr ein Logarithmus? Zehner, Zweier, natrlicher...
int Axis::genLogTicks(DWORD flags,int intervals) {
  ticks.clear();	// Ruft Destruktoren auf
  ensureMin(intervals,1);
  Float dx=Float(::log10(dispRange.e/dispRange.a));
  dx=Float(::floor(dx/intervals));	// 10 oder 100 oder 1000 ?
  ensureMin(dx,1.f);		// mindestens 10
  dx=Float(::pow(10.f,dx));
  Float p;
  if (flags&1<<Axis::bIntegralB) {
    p=dispRange.a=roundToLog(dispRange.a,-10);
  }else{
    p=roundToLog(dispRange.a,10);
  }
  ticks.reserve(intervals+1);
  int i=0;
  for(;i<=intervals;p*=dx){
    if (!(flags&1<<Axis::bIntegralE) && p>dispRange.e) break;
    new(&ticks[i++]) Tick(*this,p);	// no string
    if (p>=dispRange.e) break;
  }
  if (flags&1<<Axis::bIntegralE) dispRange.e=p;
  ticks.commit_undefined(i);
  return i;
}


// Den Bereich trotz Autoscale so festlegen, dass bei bekannter Auflsung
// (des A/D-Wandlers) nicht das Quantisierungsrauschen den bestimmenden Teil ausmacht,
// d.h. das Quantisierungsrauschen unter bspw. 10 Pixel zu drcken
// (Nicht bei Zeitachse anzuwenden!)
void DisplayRange::usefulRange(int minimumOfTicks) {
  if (!(aufloes>0)) return;	// Ohne gegebene Auflsung nichts zu machen
  if (!(minimumOfTicks>2)) return;
  Float l=minimumOfTicks*aufloes;
  if (delta()>=l) return;
// Zu wenig dargestellter Bereich:
// Bei unterschiedlichen Vorzeichen von a und e um die Mitte herum
  if (a<0 && e>0) {
    if (max(-a,e)<=l/2) {
      a=-l/2; e=l/2;
    }else if (e>-a) a=e-l;
    else e=a+l;
  }else if (e<=0) {	// Bei gleichen Vorzeichen zur Null hin
    if (a<-l) e=a+l;	// bleibt negativ;
    else {a=-l; e=0;}
  }else if (e>l) a=e-l;
  else{
    a=0; e=l;	// hufigster Fall!
  }
  Float _e=a+minimumOfTicks*aufloes;
  if (e<_e) e=_e;
}

/********
 * Axis *
 ********/

// Schwere Entscheidung: Soll dieser Konstruktor die Achse in p.axes einfgen oder nicht??
Axis::Axis(PlotXY&p,const AXISATTR&a):
 pl(p),		// Zeiger kopieren
 AXISATTR(a),	// Daten schaufeln, array_ptr<> raubkopieren, Strings kopieren
 cfore(0),
 ctitl(0),
 ptick(Pen::create(1,cfore)),
 ftick(Font::create(TEXT("Arial"),14)),
 ftitl(Font::create(TEXT("Arial"),15,Font::bold,a.side&1?a.side&2?-900:900:0)),
 scale(*(a.side&1<<bLog?static_cast<Scale*>(new LogScale):new LinScale)) {
// sichtbar und Platz beanspruchend wird eine Achse erst durch Verwendung durch einen sichtbaren Trace!
// TODO: Mindestens 1 X-Achse und 1 Y-Achse mssen pro (sichtbarer) Plotarea sichtbar bleiben
 if (name.empty()) name=isTime()?TEXT("time"):TEXT("value");
 makeTitle();
}

void Axis::distributeTicks() {
}

// liefert true wenn die angebotenen Achsenattribute hinreichend gut passen
bool Axis::operator==(const AXISATTR&a) const{
 return !((side^a.side)&0x7D)			// X/Y sowie Typ muss passen
 && (a.name.empty() || !StrCmpI(name,a.name))	// Wenn Name gegeben muss Name passen (Gro/Kleinschreibung egal)
 && (a.unit.empty() || !StrCmp(unit,a.unit))	// dasselbe fr Einheit (Gro/Kleinschreibung NICHT egal!)
 && (a.iaxis<0 || 
 iaxis==a.iaxis)		// bei Achsenwunsch muss Nummer passen
 && (a.iplotxy<0 || 
 iplotxy==a.iplotxy);	// bei Plotbereichswunsch muss Zeile/Spalte passen
}

int YAxis::needSpace(Margin<>&need) {
 updateScaling(false);	// generate labels
 HDC dc=pl.getdc();
 SelectFont(dc,ftick);
 SIZE sz={0};	// Auf Fehlersuche
 int h;
 if (isDigital()) {
  GetTextExtentPoint32(dc,TEXT("0"),1,&sz);
  need.r=sz.cx+metrics.ticklen;
  SelectFont(dc,ftitl);
  GetTextExtentPoint32(dc,name,name.length(),&sz);
  need.l=sz.cx;	// Textbreite (links)
  h=sz.cy;
  need.l=sz.cy>>1;
  need.r=sz.cy-need.t;
 }else{
  Rect<>ticksRect(allTicksRect());	// Umhllende aller Textrechtecke
  int hh=pl.wh_cr(1,row);	// derzeitige Hhe (darauf beziehen sich die Textrechtecke!)
  col=pl.irc(*this);	// links oder rechts
  row=pl.icr(*this);	// niemals am Rand
  need.l=max(8,-ticksRect.top);		// was oben bersteht als Margin.top
  need.r=max(8, ticksRect.bottom-hh);	// was unten bersteht als Margin.bottom
  int w=ticksRect.width()+metrics.ticklen;
  if (iaxis==1) DEBUG(("w=%d",w));
  h = 20;
  if (!title.empty()) {
   SelectFont(dc,ftitl);
   SIZE szt={0};
   GetTextExtentPoint32(dc,title,title.length(),&szt);
   w+=szt.cy;	// Titelhhe dazu
   toMax(h,int(szt.cx));
   int kk=need.l+hh+need.r;	// Hhe-mit-Rand zurzeit
   if (szt.cx>kk) {		// Titeltext lnger als das?
    kk=szt.cx-hh;		// Neuer berstand
    int ko=kk>>1;		// halbieren
    ensureMin(need.l,ko);	// berstand gleichmig verteilen
    ensureMin(need.r,kk-ko);	// BUG: Platz verschwendend bei stark asymmetrischem ticksRect
   }
  }
  need.t=side&1<<bBottomRight?0:w;
  need.b=side&1<<bBottomRight?w:0;
 }
 pl.releasedc();
 return h;
}

int XAxis::needSpace(Margin<>&need) {
 updateScaling(false);	// generate labels
 HDC dc=pl.getdc();
 SelectFont(dc,ftick);
 SIZE sza={0},sze={0};
 TCHAR buf[64];
 int xa=0,xe=0;
 if (!ticks.empty()) {
   Tick&ta=ticks.front();
   int l=FloatStr(ta.p,buf,elemof(buf));
   if (!GetTextExtentPoint32(dc,buf,l,&sza))
	DEBUG(("GetLastError=%u",GetLastError()));
   xa=0-scale(ta.p);
   Tick&te=ticks.back();
   l=FloatStr(te.p,buf,elemof(buf));
   ASSERT(GetTextExtentPoint32(dc,buf,l,&sze));
   xe=scale(te.p)-pl.wh(*this);	// negativ
  }else{
   int l=FloatStr(dispRange.a,buf,elemof(buf));
   ASSERT(GetTextExtentPoint32(dc,buf,l,&sza));
   l=FloatStr(dispRange.e,buf,elemof(buf));
   ASSERT(GetTextExtentPoint32(dc,buf,l,&sze));
  }
  size_t index_anfang = side>>1&2;
  need[index_anfang]=xa+(sza.cx>>1);
  need[index_anfang^2]=xe+(sze.cx>>1);
  int w=sza.cx+sze.cx;
  int h=metrics.ticklen+sza.cy;
  if (!title.empty()) {
   SelectFont(dc,ftitl);
   SIZE sz;
   GetTextExtentPoint32(dc,title,title.length(),&sz);
   ensureMin(w,(int)sz.cx);
   h+=sz.cy;	// Texthhe dazu
 }
 need.t=side&1<<bBottomRight?0:h;
 need.b=side&1<<bBottomRight?h:0;
 pl.releasedc();
 return w-need.getW();		// Mindestbreite fr alle Plotareas
}

bool Axis::anyTrace(bool(Trace::*testfunc)()const, bool runwhile) const{
  for (traces_t::const_iterator it=traces.begin(); it!=traces.end(); it++) {
    if (((*it)->*testfunc)()!=runwhile) return !runwhile;
  }
  return runwhile;
}
/* Nicht richtig ausprogrammiert:
updateScaling kann aufgerufen werden:
* wenn sich bei AutoExpand Bereichsgrenzen vergrern
* wenn sich bei AutoShrink Bereichsgrenzen verkleinern
* bei Festlegung einer Bereichgrenze (und Lschen von AutoExpand und AutoShrink)
* bei nderung von ShowZ, IntegralB, IntegralE, CenterZ
* beim Zoomen (faktisch dasselbe wie Bereichsgrenzen ndern)
* beim Verndern der Plotbereichsgre (KEINE nderung von Bereichsgrenzen!)
* beim Verndern von Traceattributen wie Linienbreite oder Marker (braucht Platz am Rand)
* beim Hinzufgen und Entfernen von Traces (kann Platz am Rand ndern)
Hier sollte dispRange unangetastet bleiben!!
*/

bool Axis::updateScaling(bool) {
 Range<Float>save(dispRange);	// tiefe Kopie
 if (/*allowShrink &&*/ dataRange.hasSpan()) dispRange.prepareFindRange();	// auf +Inf/-Inf stellen
 if (dispRange.expand(dataRange)	// Liefert <true> bei Bereichsvernderung
 || ticks.empty()) {
  HDC dc=pl.getdc();
  traces_t::const_iterator it;
  if (!isDigital()) {
   Float W = Float(pl.wh(*this)-metrics.tickwidth);
      // Achtung - Bei leerem "traces"-Array wrde Math.min +Infinity liefern
   int w=metrics.fontheight+metrics.gap;	// fr Y-Achsen
   Float aufloes = FINF;
   for (it=traces.begin(); it!=traces.end(); it++) {
    DATA*data = (*it)->getData(side&1);
    if (data) ensureMin(aufloes,data->precis());
   }
   if (!my::isfinite(aufloes)) aufloes=0;
   dispRange.aufloes = aufloes;
   if (isLog()) {
    genLogTicks(side,lrint(W/w));
   }else{
    if (!isTime() && !anyTrace(&Trace::editable))
	dispRange.usefulRange(lrint(W/metrics.steppx));	// dispRange ggf. nach oben erweitern
    if (!(side&1)) {	// X-Achse: Textbreiten messen
     genLinTicks(side,20);	// 20 = pauschale Menge an Intervallen
     SelectFont(dc,ftick);
     w=0;
     for (ticks_t::iterator it=ticks.begin();it!=ticks.end(); it++) {
      ensureMin(w,it->getRect().width());
     }
     w+=metrics.gap*2;	// mehr Platz bei X-Achsen
    }
    genLinTicks(side,lrint(W/w));	// verndert dispRange!!
   }
    //int sp=needSpace(false);
    //if (sp>metrics.lastSpaceNeeded) plot->update(1);	// "arrange" auslsen
    //if (sp<metrics.lastSpaceNeeded-metrics.gap) plot->update(1);
  }
  memset(&padding,0,sizeof padding);
  padding[side&1^2] = 1;	// 1 Pixel fr Tick-Linienbreite rechts oder unten
  for(it=traces.begin(); it!=traces.end(); it++) (*it)->measure(padding);
  padding[side&1^1] = 0;	// Falsches Padding entfernen
  padding[side&1^3] = 0;
  pl.releasedc();
 }
// bool changed=dispRange!=save;
 int A=padding.at(side&1),
     E=pl.wh(*this)-padding.at(side&1|2);
 if (side&1<<bReversed) swap(A,E);
 Float a=dispRange.a, e=dispRange.e;
 if (side&1<<bCenterZ) {
  if (a>-e) a=-e;
  else if (e<-a) e=-a;
 }
 scale.make(a,e,A,E);
 buildTickFormat();
 if (iaxis==1) 
	DEBUG(("Achse %d updateScaling %u Ticks, dispRange=(%g,%g), dataRange=(%g,%g)",
	iaxis,ticks.size(),dispRange.a,dispRange.e,dataRange.a,dataRange.e));
/* if (changed) {
  pl.flags|=1<<PlotXY::bArrangeAll;
  pl.invalidate(/ *all* /); 
 }else*/ pl.invalidate(*this);
 shrink_countdown = shrink_time*100;	// 1 s
 return true/*changed*/;
}

Range<Float> Axis::allTracesRange() const{
 Range<Float> r;	// initialisiert mit Nullen, d.h. 0 ist Teil des Bereichs
 if (side&1<<bLog || !(side&(1<<bShowZ|1<<bCenterZ))) r.prepareFindRange();
 for (traces_t::const_iterator it=traces.begin(); it!=traces.end(); it++) {
  r.expand((*it)->getRange(side&1));	// Von jedem Trace den Bereich (aus dem Cache) nehmen
 }
 if (!r.hasSpan()) r.init(0,10);
 return r;
}

bool Axis::checkAutorange() {
 if (!isDigital() && side&1<<bAutoExpand) {
  Range<Float> r(allTracesRange());	// liefert ungltigen Range wenn kein Trace vorhanden
  bool expand = r.hasSpan() && !(r<=dispRange);	// Wenn Bereich an mindestens einer Seite nicht mehr hineinpasst
  if (expand) {
   dataRange=r;
   updateScaling(false);	// da verndert sich mglw. der ganze Plot!
  }
  return expand;
 }
 return false;
}

void Axis::checkRange(bool allowShrink) {
  Range<Float> r;	// initialisiert mit Nullen, d.h. 0 ist Teil des Bereichs
  if (isDigital()) r.init(0,1);	// Digital: Fester Bereich
  else if (side&1<<bAutoExpand || allowShrink && side&1<<bAutoShrink) {
    r = allTracesRange();
    if (!r.hasSpan(precis)) return;
    if (allowShrink || dataRange!=r) {
      dataRange=r;	// auf Skale anwenden
      updateScaling(allowShrink);	// TODO: runden + verzgertes Shrinken
    }
  }
}

void Axis::makeTitle() {
 TCHAR buf[40];
 wnsprintf(buf,elemof(buf),unit.empty()||unitToTickFormat()?TEXT("%s"):TEXT("%s in %s"),name.c_str(),unit.c_str());
 title=buf;
}
/*
TCHAR*Axis::axistext() const {
  TCHAR*ret=new TCHAR[256];
  int l=axistext(ret,256);
  realloc(ret,(l+1)*sizeof(TCHAR));	// shrink-only, should always succeed
  return ret;
}
*/
bool Axis::unitToTickFormat() const{
 return unit.length()==1 && !IsCharAlphaNumeric(unit[0]);	// Symbol, hier ''
}

void Axis::buildTickFormat() {
 TCHAR*tf=tickformat;
 *tf++='%';
 Float absmax=max(fabs(dispRange.a),fabs(dispRange.e));
 char fmt='f';
 int prec;
 if (side&1<<bLog) {
  Float ratio=dispRange.e/dispRange.a;
  ASSERT(ratio>0);
  if (ratio<1) ratio=1/ratio;
  if (ratio>100) {
   fmt='g';
   prec=1+lrint(ratio);
  }
 }else{
  Float diff=dispRange.delta();
  ASSERT(diff>0);
// TODO: Differenzielle Tickbeschriftung fr starke Vergrerungen oder Zeitachsen mit Zeitstempel
  prec=1-lrint(log10(diff));	// diff=1: prec=1, diff=10: prec=0 usw.
  if (-3<=prec && prec<=6) {
   if (prec<0) prec=0;
  }else{
   fmt='g';	// Exponentialdarstellung
   prec=1+lrint(log10(absmax/diff));
  }
 }
 *tf++='.';
 if (prec) *tf++='0'+prec;
 *tf++=fmt;
 if (unitToTickFormat()) *tf++=unit[0];	// Symbol, hier ''
 *tf=0;
}

int Axis::FloatStr(Float v, TCHAR*buf,int blen) const{
 int l=_sntprintf(buf,blen,tickformat,v);
 TCHAR*p=_tcschr(buf,'.');
 if (p) *p=pl.systemsettings.sDecimal[0];
 return l;
}

cstr<> Axis::FloatStr(Float v) const{
 TCHAR buf[16];
 FloatStr(v,buf,16);
 return buf;
}

Rect<> Axis::allTicksRect() {
 Rect<> r;
 r.prepareExpand();
 for (ticks_t::iterator it=ticks.begin();it!=ticks.end();it++) r+=it->getRect();
 if (r.left>r.right || r.top>r.bottom) r.zero();
 return r;
}

/*
void Axis::checkNeed() {
  bool need=false;
  for (traces_t::iterator it=traces.begin();it!=traces.end();it++) {
    if ((*it)->visible()) {need=true; break;}
  }
  setVisible(need);
}

int Axis::textWidth(const TCHAR*s, int l) const{
  return pl.textWidth(s,l);
}

int Axis::getTitle(TCHAR*buf,int buflen) const{
 return _sntprintf(buf,buflen,
	unit.empty()?TEXT("%s"):TEXT("%s in %s"),name.c_str(),unit.c_str());
}
*/
void YAxis::paint(HDC dc) {
 if (!visible()) return;
 int w=pl.wh_cr(*this), h=pl.wh(*this);
 if (ticks.size()) {
  int	x1=side&2?0:w,				// Linienstart
	x2=x1+negIf(!(side&2),metrics.ticklen),	// Linienende
	x3=x2+negIf(!(side&2),metrics.gap);	// Textstart
  SelectPen(dc,ptick);
  SelectFont(dc,ftick);
  Metric metric(dc);
  int textmiddle=(metric.ascent-metric.intlead)>>1;
  SetTextColor(dc,cfore);
  SetTextAlign(dc,side&2?TA_LEFT|TA_BASELINE:TA_RIGHT|TA_BASELINE);	// Ausrichtung
  for (ticks_t::iterator it=ticks.begin(); it!=ticks.end(); it++) {
   int y=it->getXY();	// getXY() makes it->text
   MoveToEx(dc,x1,y,0);
   LineTo(dc,x2,y);
   TextOut(dc,x3,y+textmiddle,it->text,it->text.length());
  }
 }
 if (!title.empty()) {
  SetTextColor(dc,ctitl);
  SelectFont(dc,ftitl);
  SetTextAlign(dc,TA_TOP|TA_CENTER);	// "hngend am Kopf"
  TextOut(dc,side&2?w:0,h>>1,title,title.length());
 }
}

void XAxis::paint(HDC dc) {
 if (!visible()) return;
 int w=pl.wh(*this), h=pl.wh_cr(*this);
 if (ticks.size()) {
  int y1=side&2?0:h;
  int y2=y1+negIf(!(side&2),metrics.ticklen);
  int y3=y2+negIf(!(side&2),metrics.gap);
  SelectPen(dc,ptick);
  SelectFont(dc,ftick);
  SetTextColor(dc,cfore);
  SetTextAlign(dc,side&2?TA_TOP|TA_CENTER:TA_BOTTOM|TA_CENTER);	// Ausrichtung
  for (ticks_t::iterator it=ticks.begin(); it!=ticks.end(); it++) {
   int x=it->getXY();	// getXY() makes it->text
   MoveToEx(dc,x,y1,0);
   LineTo(dc,x,y2);
   TextOut(dc,x,y3,it->text,it->text.length());
  }
 }
 if (!title.empty()) {
  SetTextColor(dc,ctitl);
  SelectFont(dc,ftitl);
  SetTextAlign(dc,side&2?TA_BOTTOM|TA_CENTER:TA_TOP|TA_CENTER);	// nur unten "stehend auf dem Fu", sonst "hngend am Kopf"
  TextOut(dc,w>>1,side&2?h:0,title,title.length());
 }
}

// Skalieren unter Beachtung von INF und NaN:
// Bei NaN wird in die Achsenmitte "skaliert", bei Inf an den jeweiligen Rand
// Der Rckgabewert liefert dann die Info, welchem Wert v hatte
char Axis::scaleInf(Float v, int&V) const{
 if (my::isfinite(v)) {	// Normale Zahl?
  V=scale(v);
  return 0;
 }
 if (v!=v) {			// NaN?
  V=pl.wh(*this)>>1;	// horizontale oder vertikale Mitte des Skalenrechtecks
  return CNAN;
 }
 char negsign=((char*)&v)[sizeof v-1]>>7;	// Vorzeichenbit achtfach
 if ((side^negsign)&4) V=0;	// Linker Rand
 else V=pl.wh(*this);		// Rechter Rand
 return negsign?-CINF:CINF;	// 127 als Infinity-Kennung liefern
}

Tick* Axis::findEditableLabel(const Point<>&pt) {
  if (ticks.size()<2) return 0;
  Tick&tf = ticks.front();
  if (tf.getRect().inside(pt)) return &tf;
  Tick&tb = ticks.back();
  if (tb.getRect().inside(pt)) return &tb;
  return 0;
}

Tick::Tick(Axis&a,Float pp,const TCHAR*t,char i):ax(a) {
 p=pp;
 text=const_cast<TCHAR*>(t);
 importance=i;
 invalidate();	// Integer-Angaben ungltig
 align=-1;	// Text von auen oder noch nicht ermittelt
}

void Tick::makeText() {
 if (text.empty()) text=ax.FloatStr(p);
}

void Tick::update() {
 if (align>=0) return;	// nichts zu tun
// 1. Aktion: Skaliertes <p> ermitteln
 xy = ax.scale(p);
// 2. Aktion: Textanker (Punkt) in der Skale ermitteln  ohne Mittenverschiebung wegen fehlendem TA_MIDDLE
 DWORD side=ax.side;
 anker.xy[side&1] = xy;
 int d = ax.metrics.ticklen + ax.metrics.gap;
 anker.xy[side&1^1] = side&2 ? d : ax.pl.wh_cr(ax)-d;
// 3. Aktion: Text ermitteln
 makeText();
// 4. Textrechteck ermitteln (Der Tick-Text ist stets horizontal)
 HDC dc=ax.pl.getdc();
 Metric metric(dc);
// 4.1 Ankerpunkt bernehmen
 ta.left=anker.x;
 ta.top=anker.y;
 align = side&1	? side&2 ? TA_LEFT|TA_BASELINE : TA_RIGHT|TA_BASELINE
		: side&2 ? TA_TOP|TA_CENTER : TA_BOTTOM|TA_CENTER;
 adjust.x = 0;	// an horizontaler Achse erforderlich wenn an jede Zahl die Einheit angehangen wird, bspw. bei 
 adjust.y = 0;
// 4.2 Textabmessungen ermitteln; r.right = Breite, r.bottom = Hhe
 if (GetTextExtentPoint32(dc,text,text.length(),(SIZE*)&ta.right)) {
// 4.3 Je nach Align Rechteck versetzen: Zuerst horizontal
// Bei X-Achse gilt TA_CENTER, bei Y-Achse TA_LEFT oder TA_RIGHT
  if (~side&1) ta.left-=ta.right>>1;	// X-Achse, also TA_CENTER
  else if (~side&2) ta.left-=ta.right;	// Y-Achse linksseitig, also TA_RIGHT
// Dann vertikal. Bei X-Achse gilt TA_TOP oder TA_BOTTOM, bei Y-Achse TA_BASELINE
  if (side&1) {	// Y-Achse, also TA_BASELINE und komplizierte Korrektur
   ta.top -= adjust.y = char((metric.ascent+metric.intlead)>>1);
  }else if (~side&2) ta.top-=ta.bottom;	// X-Achse oben: TA_BOTTOM
// 4. Breite/Hhe in rechts/unten umwandeln
  ta.right += ta.left;
  ta.bottom += ta.top;
 }else SetRectEmpty(ta);
 ax.pl.releasedc();
}

void setMenuTitle(HMENU m, UINT id, const TCHAR*text) {
 if (text && *text) {
  TCHAR s[64],buf[64];
  GetMenuString(m,id,s,elemof(s),0);
  wnsprintf(buf,elemof(buf),TEXT("%s: %s"),s,text);
  ModifyMenu(m,id,0,id,buf);
 }
}

void Axis::prepareMenu(HMENU m) {
 ASSERT(bAutoExpand==8);
 BYTE b=BYTE(side>>8);
 for (BYTE id=16; id<24; id++,b>>=1) {	// 8 Items in passender Reihenfolge hintereinander
  if (b&1) CheckMenuItem(m,id,MF_CHECKED);
 }
 setMenuTitle(m,11,name);
 setMenuTitle(m,12,unit);
 TCHAR s[64];
 LoadString(0,20,s,elemof(s));	// "links/oben/rechts/unten"
 const TCHAR*p=s;
 b=BYTE(side)&3;		// Auswahl des richtigen Worts
 if (b) do p+=lstrlen(p)+1; while(--b);
 setMenuTitle(m,9,p);		// "Seite"
 if ((side>>bVertical^side>>bReversed)&1) CheckMenuItem(m,10,MF_CHECKED);

// MENUITEMINFO mii;
// mii.cbSize=sizeof mii;
// mii.fMask=MIIM_STATE;
// GetMenuItemInfo(m,17,false,
}

void Axis::onRButtonDown(Point<>&pt) {
 DWORD pos=GetMessagePos();
 HMENU m=LoadMenu(0,MAKEINTRESOURCE(20));
 HMENU sub=GetSubMenu(m,0);
 prepareMenu(sub);
 byte sel=byte(TrackPopupMenu(sub,
	TPM_RIGHTBUTTON|TPM_RETURNCMD,GET_X_LPARAM(pos),GET_Y_LPARAM(pos),
	0,pl.hPlot,0));
 if (16<=sel && sel<24) {
  BYTE&b=((BYTE*)&side)[1];
  b^=1<<(sel-16);
 }else switch (sel) {
  case 9: side^=1<<bBottomRight; break;	// oben/unten wechseln
  case 10: side^=1<<bReversed; break;	// Anstiegsrichtung wechseln
  case 99: CreateDialogParam(0,MAKEINTRESOURCE(20),pl.hPlot,BereichDlgProc,(LPARAM)this);
 }
 DestroyMenu(m);
}

bool getWindowFloat(HWND Wnd, Float&v) {
 TCHAR s[32];
 GetWindowText(Wnd,s,elemof(s));
 TCHAR*p=StrChr(s,*PlotXY::systemsettings.sDecimal);
 if (p) *p='.';
 return _stscanf(s,sizeof v>4?TEXT("%lf"):TEXT("%f"),&v)==1;
}

INT_PTR _stdcall Axis::BereichDlgProc(HWND Wnd, UINT msg, WPARAM wParam, LPARAM lParam) {
 Axis*ax=(Axis*)GetWindowLongPtr(Wnd,DWLP_USER);
 switch (msg) {
  case WM_INITDIALOG: {
   SetWindowLongPtr(Wnd,DWLP_USER,lParam);
   SetTimer(Wnd,1,200,0);	// Nicht schn aber funktionell
  }return TRUE;

  case WM_COMMAND: {
   switch (wParam) {
    case MAKELONG(IDCANCEL,BN_CLICKED): DestroyWindow(Wnd); return TRUE;
    case MAKELONG(16,BN_CLICKED): ax->side^=1<<bAutoExpand; break;
    case MAKELONG(17,BN_CLICKED): ax->side^=1<<bAutoShrink; break;
    case MAKELONG(18,EN_KILLFOCUS):
    case MAKELONG(19,EN_KILLFOCUS):
    case MAKELONG(20,EN_KILLFOCUS):
    case MAKELONG(21,EN_KILLFOCUS): {
     HWND hEdit=(HWND)lParam;
     if (Edit_GetModify(hEdit)) {
      Edit_SetModify(hEdit,FALSE);
      switch (LOBYTE(wParam)) {
       case 18: ax->shrink_time=BYTE(GetDlgItemInt(Wnd,18,NULL,FALSE)/100); break;
       case 19: ax->shrink_perc=BYTE(GetDlgItemInt(Wnd,19,NULL,FALSE)); break;
       case 20: getWindowFloat(hEdit,ax->dispRange.a); break;
       case 21: getWindowFloat(hEdit,ax->dispRange.e); break;
      }
     }
    }
   }
  }break;

  case WM_TIMER: {
   BOOL ena=ax->side&3<<bAutoExpand?FALSE:TRUE;
   EnableWindow(GetDlgItem(Wnd,20),ena);
   EnableWindow(GetDlgItem(Wnd,21),ena);
   CheckDlgButton(Wnd,16,ax->side&1<<bAutoExpand?BST_CHECKED:0);
   CheckDlgButton(Wnd,17,ax->side&1<<bAutoShrink?BST_CHECKED:0);
   HWND hFocus=GetFocus(), hTarget;
   hTarget=GetDlgItem(Wnd,18);
   if (hFocus!=hTarget) SetDlgItemInt(Wnd,18,ax->shrink_time*100,FALSE);
   hTarget=GetDlgItem(Wnd,19);
   if (hFocus!=hTarget) SetDlgItemInt(Wnd,19,ax->shrink_perc,FALSE);
   hTarget=GetDlgItem(Wnd,20);
   if (hFocus!=hTarget) SetWindowText(hTarget,ax->FloatStr(ax->dispRange.a));
   hTarget=GetDlgItem(Wnd,21);
   if (hFocus!=hTarget) SetWindowText(hTarget,ax->FloatStr(ax->dispRange.e));
  }break;
 }
 return FALSE;
}
