/* axis+trace fabrication parser
Uses the window text to create+entitle axes, units, colors, etc.
This way, axes and plots can be defined as resource string!
Main entities start with '\1' upto '\7',
whereas sub-entities start with '\x08' upto '\x1F', so tabs and newlines are never part of an entity.
All entities end on codes < '\x20'.
Main entities:
 '\x01'	'\001'	global attributes
 '\x02'	'\002'	X axis
 '\x03'	'\003'	Y axis
 '\x04'	'\004'	Trace, or XY graph
 '\x05'	'\005'	Cursor
 '\x06'	'\006'	Annotation
 '\x07'	'\007'	?							'\a'
Sub-entities:	1	2	3	4	5	6	7
 '\x08'	'\010'		name	name	name				'\b'
 '\x09'	'\011'		unit	unit	linesty				'\t'
 '\x0A'	'\012'		format	format	marker				'\r'
 '\x0B'	'\013'		column	row	fillTo				'\v'
 '\x0C'	'\014'	C:back	C:axis	C:axis	C:line				'\f'
 '\x0D'	'\015'	C:area	C:text	C:text	C:fill				'\n'
 '\x0E'	'\016'	C:lgnd	C:grid	C:grid	C:mark
 '\x0F'	'\017'	font	font	font	font	font	font
 '\x10'	'\020'				xaxis
 '\x11'	'\021'				yaxis
 '\x12'	'\022'
 '\x13'	'\023'
 '\x14'	'\024'
 '\x15'	'\025'
 '\x16'	'\026'
 '\x17'	'\027'
 '\x18'	'\030'
 '\x19'	'\031'
 '\x1A'	'\032'
 '\x1B'	'\033'								'\e'
 '\x1C'	'\034'
 '\x1D'	'\035'
 '\x1E'	'\036'
 '\x1F'	'\037'
C:color (to be defined, system color index possible, alpha possible)
*/
#include "cplotxy.h"
#include <tchar.h>

// Teilstringterminator < 0x20 finden. Damit sind Newlines und Tabs im String verboten.
static const TCHAR*strend(const TCHAR*s) {
 while ((unsigned)*s>=' ') ++s;
 return s;
}

// Vierkanal-Oszifarben sind zumeist gelb-blau-rot-grn
static const COLORREF DefColor[4]={RGB(192,192,0),RGB(0,128,255),RGB(255,65,0),RGB(0,192,0)};

class WindowTextParser{
 PlotXY&pl;
public:
 char item,subitem;
 AXISATTR axisattr[2];
 TRACEATTR traceattr;

 WindowTextParser(PlotXY&p):pl(p) {
  axisattr[0].side=Axis::BOTTOM|/*1<<Axis::bTime|*/1<<Axis::bAutoExpand|1<<Axis::bAutoShrink|1<<Axis::bShowB|1<<Axis::bShowE|1<<Axis::bShowZ|1<<Axis::bVisible;
  axisattr[1].side=Axis::LEFT|1<<Axis::bAutoExpand|1<<Axis::bAutoShrink|1<<Axis::bVisible|1<<Axis::bShowB|1<<Axis::bShowE|1<<Axis::bShowZ;
 }

 void begin(char newitem) {
  item=newitem;
  subitem=0;	// Name
  switch (newitem) {
   case 4: {
    traceattr.ixaxis=-1;	// Achse automatisch
    traceattr.iyaxis=-1;
    traceattr.iplotx=-1;	// Plotbereich egal
    traceattr.iploty=-1;
    traceattr.tcolor=DefColor[traceattr.itrace>=0?traceattr.itrace&3:0];
   }break;
  }
 }

 void beginSub(char newsub) {
  subitem=newsub;
 }

 void end() {
  switch (item) {
   case 2:
   case 3: {
    AXISATTR&aa=axis();
    pl.insertAxis(aa);
    aa.side^=1<<Axis::bBottomRight;	// links/rechts oder oben/unten wechseln
    aa.name=0;
    aa.unit=0;
    aa.iaxis++;
  }break;
   case 4: {
    TRACEATTR&ta=traceattr;
    pl.insertTrace(ta);
    ta.name=0;
    ta.itrace++;
   }
  }
 }

 AXISATTR&axis() {return axisattr[item&1];}

 TCHAR*newStr(const TCHAR*&s,const TCHAR*e) {
  TCHAR*ret=0;
  size_t l=e-s;
  if (l) {
   ret=new TCHAR[l+1];
   memcpy(ret,s,l*sizeof(TCHAR));
   ret[l]=0;
  }
  s=e;
  return ret;
 }

 int getInt(const TCHAR*&s,const TCHAR*e) {
  int ret=0;
  size_t l=e-s;
  if (l) {
   TCHAR*eos;
   ret=_tcstoul(s,&eos,0);
   if (eos!=e) {	// Dann war es keine Zahl
// Infrage kommende Objektliste nach Namen absuchen
    for (PlotXY::axes_t::const_iterator ia=pl.axes.begin(); ia!=pl.axes.end(); ia++) {
     cstr<>&n=ia->get()->name;
     if (n.length()==(int)l && !memcmp(n,s,l*sizeof(TCHAR))) {	// BUG: Case-insensitiver Vergleich muss her, wenn schon
      ret=(int)(ia-pl.axes.begin());
      break;
     }
    }
   }
   s=e;
  }else if ((unsigned)*s<0x20) {
   ret=*s++;
  }
  return ret;
 }
};

// Text aus Ressource verarbeiten:
// Die verschieden funktionalen Teilstrings haben folgende Präfix-Zeichen:
// '\2' = X-Achse folgt: \20Name\21Einheit\22Format\23Spalte\24Attr
// '\3' = Y-Achse folgt: \20Name\21Einheit\22Format\23Zeile \24Attr
// '\4' = Trace folgt: \20Name\30X-Achse\31Y-Achse\32Farbe\33Linienbreite\34Linienart\35Verbinder\36Marker
// X-Achsen gelangen per Vorgabe abwechselnd unten-oben; Spalte = im tabellarischen Plotbereich nebeneinander
// Y-Achsen gelangen per Vorgabe abwechselnd links-rechts; Zeile = im tabellarischen Plotbereich übereinander
// Als Zahlen vorgebbare Parameter sind Zeichen, wenn <0x20, Dezimalzahlen oder (Achsen-)Namen
void PlotXY::onSetText(const TCHAR*s) {
 WindowTextParser parser(*this);
 for(;;) {
  TCHAR c=*s;
  if (!c) {parser.end(); break;}
  if ((unsigned)c<' ') {
   ++s;
   if (c<8) {
    parser.end();
    parser.begin(char(c));
   }else parser.beginSub(char(c));
  }
  const TCHAR*e=strend(s);	// Zwischen s und e liegt jetzt der Teilstring. Bei s==e kann das ein char-Numeral oder ein Leerstring sein
  switch (parser.item) {
   case 2:			// X-Achse
   case 3: switch (parser.subitem) {	// Y-Achse
    case 0:
    case 010: parser.axis().name=parser.newStr(s,e); break;
    case 011: parser.axis().unit=parser.newStr(s,e); break;
    case 013: {
     AXISATTR&ax=parser.axis();
     int i=parser.getInt(s,e);
     if (ax.iplotxy!=i) {
      ax.iplotxy=char(i);
      if (ax.side&1) ax.side&=~2;	// Y-Achse links beginnen
      else ax.side|=2;			// X-Achse unten beginnen
     }
    }break;
    default:
    TCHAR*str=parser.newStr(s,e);
    DEBUG(("parser.item=%X, parser.subitem=%X, arg=%"SST,parser.item,parser.subitem,str));
    if (str) delete[]str;
   }break;
   case 4: switch (parser.subitem) {	// Trace
    case 0:
    case 010: parser.traceattr.name=parser.newStr(s,e); break;
    case 020: parser.traceattr.ixaxis=char(parser.getInt(s,e)); break;
    case 021: parser.traceattr.iyaxis=char(parser.getInt(s,e)); break;
   }break;
   default: DefWindowProc(hPlot,WM_SETTEXT,0,(LPARAM)parser.newStr(s,e)); break;	// Plotname
  }
 }
}

