#pragma once
/* A fixed and a growable container that never throw exceptions.
   Compatible with MSVC6. Leider nur mit /Ob0.
   heha, 240215
 */

#include <malloc.h>	// realloc, free; _alloca
#include <string.h>	// memcpy
#pragma intrinsic(memcpy,memset)
#include <windows.h>	// lstrlen, lstrcpy
#include <stddef.h>	// ptrdiff_t

template<class T>inline void swap(T&x,T&y) {T t=x; x=y; y=t;}

#undef min
template<class T>inline const T&min(const T&a, const T&b) {return a<b?a:b;}

#undef max
template<class T>inline const T&max(const T&a, const T&b) {return a>b?a:b;}
// etwa wie CopySign
template<class T> T negIf(bool e,T v) {return e?-v:v;}

// Ich bin mir nicht sicher ob der Compiler const T& zu T optimiert bei kleinen T.
// Um als 2. Parameter eine Konstante oder einen Ausdruck angeben zu knnen, erwartet C++ sicherlich T.
template<class T> bool ensureMin(T&a,T b) {if (a>=b) return false; a=b; return true;}
template<class T> bool ensureMax(T&a,T b) {if (a<=b) return false; a=b; return true;}
enum limit_t{
 WAS_BELOW=1,
 WAS_ABOVE=2,
};
template<class T> limit_t ensureMinMax(T&a,T lo,T hi) {
  return limit_t(char(ensureMin(a,lo)) | char(ensureMax(a,hi))<<1);
}
template<class T> T limit(T a,T lo,T hi) {
  return min(max(a,lo),hi);
}

// Dasselbe wie ensureMin/ensureMax aber mit einer anderen Intention: Das Verndern als Normalfall
template<class T> bool toMin(T&a,T b) {return ensureMax(a,b);}
template<class T> bool toMax(T&a,T b) {return ensureMin(a,b);}

// Placement-New: Ignore size and return passed pointer.
// After placement new is done, the (single) constructor is called.
inline void*operator new(size_t,void *P){return P;}

// logarithmus binary: Hchstwertiges Bit ermitteln, normalerweise mit count_leading_zeroes()
// Das Gegenteil ist 1UL<<n
inline char lb(size_t z) {
 char r=-1;
 while(z>=256) {z>>=8; r+=8;}	// Ganze Bytes schieben (gut fr 8-Bit-Controller)
 while(z) {z>>=1; r++;}		// Bis die letzte Null 'raus ist (in 1 Byte zu implementieren)
 return r;
}

// Round up to the next highest power of 2
inline size_t nextpow2(size_t z) {
 z--;z|=z>>1;z|=z>>2;z|=z>>4;z|=z>>8;z|=z>>16;
#ifdef _WIN64
 z|=z>>32;
#endif
 return z+1;
}

// Standard-Allokator
template<class T> class allocator{
public:
 typedef size_t size_type;
 typedef ptrdiff_t difference_type;
 typedef T*pointer;
 typedef const T*const_pointer;
 typedef T&reference;
 typedef const T&const_reference;
 typedef T value_type;
 pointer address(reference x) const {return &x;}
 const_pointer address(const_reference x) const {return &x;}
 allocator() {}
 allocator<T>& operator=(const allocator<T>) {}
 pointer allocate(size_type n, const void*) {return new[n]T;}
 void deallocate(pointer p,size_type) {delete[]p;}
 void construct(pointer p, const T&o) {new(p)T(o);}	// placement-new + copy constructor
 void destroy(pointer p) {p->~T();}			// call destructor but not free memory
 size_type max_size() const {return (1<<31)/sizeof(T);}
};

